from threading import Thread

from _packet import _Packet

from .utils import *

from time import sleep



class Guidance( Thread ):

    def __init__( self, config ):

        super().__init__( self, daemon=True )

        self.packet = None
        self.header = config['header']
        self.cf     = config['scf'].cf

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )

    
    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=1 )

        packet = self.packet

        packet._enroll_receiver( 12*4, self.header )

        thread = Thread( target=packet._recvfrom, args=(), daemon=True )

        thread.start()


    def run( self ):

        pos = self.cf.pos
        vel = self.cf.vel
        acc = self.cf.acc
        att = self.cf.rpy

        acc_cmd = self.cf.command

        if self.packet is not None:

            packet = self.packet

            packet._enroll( 3, self.header )

        while self.AllGreen:

            packet.TxData[0:3] = acc_cmd

            packet._sendto()

            pos[:] = packet.RxData[0:3]
            vel[:] = packet.RxData[3:6]
            acc[:] = packet.RxData[6:9]
            att[:] = packet.RxData[9: ]

            sleep( 0.01 )


from serial import Serial

from numpy import zeros
from numpy import frombuffer
from numpy import float32, uint8
from numpy import ndarray

_FLOAT=4



class _Packet( Serial ):

    def _enroll( self, size: int, header: ndarray ):

        self.TxData = zeros( size, dtype=float32 )
        self.header = ( header.astype( uint8 ) ).tobytes()

    
    def _enroll_receiver( self, size: int, header: ndarray ):

        self.RxData   = zeros( size, dtype=float32 )
        self.RxHeader = ( header.astype( uint8 ) ).tobytes()
        self.RxBfsize = size * _FLOAT


    def _sendto( self ):

        buffer = self.header + self.TxData.tobytes()

        self.write( buffer )

    
    def _recvfrom( self ):

        RxData   = self.RxData
        RxHeader = self.RxHeader
        size     = self.RxBfsize         ## float

        hdrf = 0

        while True:

            if self.readable():

                if ( hdrf == 2 ):

                    data = self.read( size )

                    RxData[:] = frombuffer( data, dtype=float32 )

                    hdrf = 0

                elif ( hdrf == 0 ):
                    data = self.read()
                    if ( data == RxHeader[0] ):
                        hdrf = 1

                elif ( hdrf == 1 ):
                    data = self.read()
                    if ( data == RxHeader[1] ):
                        hdrf = 2

                else:
                    hdrf = 0



from .smoother import smooth_command

from numpy import array, zeros

from time import sleep

from constants import Kp, Kd, g



def hover( cf, T, dt=0.1 ):

    cur     = zeros(6)
    des     = zeros(6)
    des_cmd = zeros(6)
    acc_cmd = zeros(3)
    P_pos   = zeros(3)
    D_pos   = zeros(3)
    care_g  = array([0,0,g])

    print( 'hover' )

    n = int( T / dt )
    t = 0

    command = cf.command

    cur[:] = cf.posvel
    posvel = cf.posvel

    des[:3] = cur[:3]
    des[3:] = 0

    cf.destination[:] = des

    for _ in range( n ):

        des_cmd[:] = smooth_command( 
            des, cur, t, int( T/2 )
        )

        P_pos[:] = des_cmd[:3] - posvel[:3]
        D_pos[:] = des_cmd[3:] - posvel[3:]

        acc_cmd[:] = 0
        acc_cmd[:] += P_pos * Kp
        acc_cmd[:] += D_pos * Kd
        acc_cmd[:] += care_g

        command[:] = acc_cmd

        t += dt

        sleep( dt )


from threading import Thread

from ..crazy import CrazyDragon

from .._packet import _Packet

from time import sleep



class Controller_TEST( Thread ):

    def __init__( self, _cf: CrazyDragon, config ):
        
        super().__init__()

        self.daemon = True

        self.packet = None
        self.header = config['header']
        self.dt     = config['dt']
        self.n      = config['n']
        self.cf     = _cf

        self.ready_for_command = True

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )

    
    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=1 )

        packet = self.packet
        packet._enroll_receiver( 3, self.header )

        thread = Thread( target=packet._recvfrom, args=(), daemon=True )
        thread.start()

    def run( self ):

        packet = self.packet
        rxData = packet.RxData

        n  = self.n
        dt = self.dt / n

        while self.ready_for_command:

            for _ in range( n ):

                print( rxData )

                sleep( dt )



from threading import Thread

from ..crazy import CrazyDragon

from .integral_loop import _dot_thrust
from .integral_loop import _thrust_clip

from .optimus_prime import _command_as_RPY
from .optimus_prime import _command_is_not_in_there

from constants import alpha

from .._packet import _Packet

from numpy import zeros, array

from time import sleep

_FLOAT = 4



class Controller( Thread ):

    def __init__( self, _cf: CrazyDragon, config ):
        
        super().__init__( self, daemon=True )

        self.packet = None
        self.header = config['header']
        self.cf     = _cf
        self.dt     = config['dt']
        self.n      = config['Hz']

        self.acc_cmd = zeros(3)
        self.command = zeros(4)
        self.thrust  = array( [alpha * 9.81], dtype=int )

        self.ready_for_command = False

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )

    
    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=1 )

        packet = self.packet

        packet._enroll_receiver( 3, self.header )

        thread = Thread( target=packet._recvfrom, args=(), daemon=True )

        thread.start()



    def init_send_setpoint( self ):
        ## commander
        commander = self.cf.commander
        ## initialize
        commander.send_setpoint( 0, 0, 0, 0 )
        self.ready_for_command = True


    def stop_send_setpoint( self ):
        ## commander
        commander = self.cf.commander
        ## stop command
        self.cf.command[:] = zeros(3)
        ## stop signal
        self.ready_for_command = False

        for _ in range( 50 ):
            commander.send_setpoint( 0, 0, 0, 10001 )
            sleep( 0.05 )

        commander.send_stop_setpoint()


    def run( self ):

        packet = self.packet
        rxData = packet.RxData

        cf        = self.cf
        commander = cf.commander
        n  = self.n
        dt = self.dt / n

        acc_cmd = self.acc_cmd
        command = self.command
        thrust  = self.thrust

        att_cur = cf.att
        acc_cur = cf.acc

        while not self.ready_for_command:
            sleep( 0.1 )

        while self.ready_for_command:

            acc_cmd[:] = rxData

            _command_is_not_in_there( acc_cmd, att_cur )

            _command_as_RPY( acc_cmd, command )

            if ( acc_cmd[2] == 0 ):
                sleep( dt )
                return

            for _ in range( n ):

                thrust[0] += _dot_thrust( command, acc_cur )

                thrust[0] = _thrust_clip( thrust[0] )

                commander.send_setpoint(
                    command[0],
                    command[1],
                    command[2],
                    thrust[0]
                )

                sleep( dt )


from threading import Thread

from ..crazy import CrazyDragon

from .imu       import IMU
from .qualisys  import Qualisys

from .._packet import _Packet

from time import sleep



class Navigation( Thread ):

    def __init__( self, _cf: CrazyDragon, config ):

        super().__init__()

        self.daemon = True

        self.packet = None
        self.header = config['header']
        self.cf     = _cf

        self.imu = IMU( config['scf'] )
        self.qtm = Qualisys( config['body_name'] )

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )


    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=1 )


    @classmethod
    def _on_pose( cls, cf: CrazyDragon, data: list ):
        
        cf.pos[:] = data[0:3]
        cf.att[:] = data[3:6]

    
    def run( self ):

        pos = self.cf.pos
        vel = self.cf.vel
        acc = self.cf.acc
        att = self.cf.att

        self.imu.start_get_vel()
        self.imu.start_get_acc()

        self.qtm.on_pose = lambda data: self.__class__._on_pose( self.cf, data )

        if self.packet is not None:

            packet = self.packet

            packet._enroll( 12, self.header )

        while self.AllGreen:

            packet.TxData[0:3] = pos
            packet.TxData[3:6] = vel
            packet.TxData[6:9] = acc
            packet.TxData[9: ] = att

            packet._sendto()

            sleep( 0.01 )



from .control import Controller_TEST

from .guidance import CommunicationCenter

from .navigation import Navigation_TEST

from time import sleep



def NAV_AND_CTR( _cf, NavConfig, CtrConfig ):

    NAV = Navigation_TEST( _cf, NavConfig )
    CTR = Controller_TEST( _cf, CtrConfig )

    NAV.start()
    CTR.start()

    while ( CTR.AllGreen and NAV.AllGreen ):
        sleep( 0.1 )


def GUIDANCE( _cf, GuiConfig ):

    CMC = CommunicationCenter( _cf, GuiConfig )

    CMC.start()

    while CMC.AllGreen:
        sleep( 0.1 )


from numpy import array
from numpy import float32, uint8
from numpy import frombuffer
from numpy import ndarray



_arr = array( [ 24, 256, 2 ], dtype=float32 )
_bit = _arr.astype( uint8 ).tobytes()
_byt = _arr.tobytes()

_fir = _byt[0:12]
_Fir = frombuffer( _fir, dtype=float32 )

print( _arr )
print( _bit )
print( _bit[1] )
print( _fir )
print( _Fir )
print( _byt[3] )


from numpy.linalg import norm

from .constants import *



K_T = alpha * Kp


def _dot_thrust(command, acc_cur):
    ## difference
    dot_acc = command[3] - norm( acc_cur )
    ## command
    dot_thr = K_T * dot_acc
    ## return
    return dot_thr


def _thrust_clip(thrust_cmd):
    ## thrust clip thrust
    if (thrust_cmd > 60000):
        thrust_cmd = 60000
    elif (thrust_cmd < 10001):
        thrust_cmd = 10001

    return int(thrust_cmd)


from threading import Thread

from ..crazy import CrazyDragon

from .._packet import _Packet

from time import sleep



class CommunicationCenter( Thread ):

    def __init__( self, _cf: CrazyDragon, config ):

        super().__init__()

        self.daemon = True

        self.packet = None
        self.Hz     = config['Hz']
        self.cf     = _cf

        self.Rxheader = config['Rxheader']
        self.Txheader = config['Txheader']

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )

    
    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=1 )

        packet = self.packet
        packet._enroll_receiver( 12, self.Rxheader )

        thread = Thread( target=packet._recvfrom, args=(), daemon=True )
        thread.start()


    def run( self ):

        packet = self.packet
        rxData = packet.RxData

        dt = 1 / self.Hz

        if self.packet is not None:

            packet = self.packet
            packet._enroll( 3, self.Txheader )

        txData = packet.TxData

        while self.AllGreen:

            packet._sendto()

            sleep( dt )



from numpy import array, zeros
from numpy import frombuffer
from numpy import float32, uint8



header = array( [4, 26], dtype=uint8 ).tobytes()
data   = array( [1,2,3], dtype=float32 )
buff   = ''
RxData = zeros( 3 )
RxBuff = zeros( 3*4 )

byte = header + data.tobytes()

print( byte )

size = 2 + 12

hdrf = 0
idxn = 0

for i in range( size ):
    _byt = byte[i]

    if ( hdrf == 2 ):
        if ( idxn < size ):
            RxBuff[idxn] = _byt
            idxn += 1

            if ( idxn == size - 2 ):
                hdrf = 0
                idxn = 0

                buff      = RxBuff.astype( uint8 ).tobytes()
                RxData[:] = frombuffer( buff, dtype=float32 )
        
    elif ( hdrf == 0 ):
        if ( _byt == header[0]):
            hdrf = 1
            print( 'pass' )
    
    elif ( hdrf == 1 ):
        if ( _byt == header[1] ):
            hdrf = 2
            print( 'pass' )

print( RxData )


from cflib.crazyflie.log import LogConfig

from .imu_setup import *



period_in_ms = 10


class IMU:

    def __init__( self, scf ):
        self.scf = scf
        self.cf  = scf.cf

        adjust_orientation_sensitivity( self.cf )
        activate_kalman_estimator( self.cf )
        reset_estimator( self.cf )


    def start_get_vel(self, period_in_ms=period_in_ms):
        log_conf = LogConfig(name="velocity", period_in_ms=period_in_ms)
        log_conf.add_variable('stateEstimate.vx', 'FP16')
        log_conf.add_variable('stateEstimate.vy', 'FP16')
        log_conf.add_variable('stateEstimate.vz', 'FP16')

        self.scf.cf.log.add_config(log_conf)
        log_conf.data_received_cb.add_callback(self.velocity_callback)
        log_conf.start()
 

    def start_get_acc(self, period_in_ms=period_in_ms):
        log_conf = LogConfig(name='acceleration', period_in_ms=period_in_ms)
        log_conf.add_variable('acc.x', 'FP16')      ## m/s^2
        log_conf.add_variable('acc.y', 'FP16')      ## m/s^2
        log_conf.add_variable('acc.z', 'FP16')      ## m/s^2

        self.scf.cf.log.add_config(log_conf)
        log_conf.data_received_cb.add_callback(self.accelerate_callback)
        log_conf.start()


    def velocity_callback(self, timestamp, data, logconf):
        self.cf.vel[0] = data['stateEstimate.vx']
        self.cf.vel[1] = data['stateEstimate.vy']
        self.cf.vel[2] = data['stateEstimate.vz']


    def accelerate_callback(self, timestamp, data, logconf):
        self.cf.acc[0] = data['acc.x'] * 9.81
        self.cf.acc[1] = data['acc.y'] * 9.81
        self.cf.acc[2] = data['acc.z'] * 9.81


from threading import Thread

from ..crazy import CrazyDragon

from .._packet import _Packet

from time import sleep



class Navigation_TEST( Thread ):

    def __init__( self, _cf: CrazyDragon, config ):

        super().__init__()

        self.daemon = True

        self.packet = None
        self.header = config['header']
        self.Hz     = config['Hz']
        self.cf     = _cf

        self.AllGreen = True

        self._on_link( config['port'], config['baud'] )


    def _on_link( self, port, baud ):

        self.packet = _Packet( port, baud, timeout=0.02 )

    def run( self ):

        pos = self.cf.pos
        vel = self.cf.vel
        acc = self.cf.acc
        att = self.cf.att

        dt = 1 / self.Hz

        if self.packet is not None:

            packet = self.packet

            packet._enroll( 12, self.header )

        txData = packet.TxData

        while self.AllGreen:

            txData[0:3] = pos
            txData[3:6] = vel
            txData[6:9] = acc
            txData[9: ] = att

            packet._sendto()

            sleep( dt )


import time

from cflib.crazyflie.log        import LogConfig
from cflib.crazyflie.syncLogger import SyncLogger



orientation_std_dev = 8.0e-3


def adjust_orientation_sensitivity(cf):
    cf.param.set_value('locSrv.extQuatStdDev', orientation_std_dev)


def activate_kalman_estimator(cf):
    cf.param.set_value('stabilizer.estimator', '2')

    # Set the std deviation for the quaternion data pushed into the
    # kalman filter. The default value seems to be a bit too low.
    cf.param.set_value('locSrv.extQuatStdDev', 0.06)


def reset_estimator(cf):
    cf.param.set_value('kalman.resetEstimation', '1')
    time.sleep(0.1)
    cf.param.set_value('kalman.resetEstimation', '0')

    # time.sleep(1)
    wait_for_position_estimator(cf)


def wait_for_position_estimator(scf):
    print('Waiting for estimator to find position...')

    log_config = LogConfig(name='Kalman Variance', period_in_ms=10)
    log_config.add_variable('kalman.varPX', 'FP16')
    log_config.add_variable('kalman.varPY', 'FP16')
    log_config.add_variable('kalman.varPZ', 'FP16')

    

    var_y_history = [1000] * 10
    var_x_history = [1000] * 10
    var_z_history = [1000] * 10

    threshold = 0.001

    with SyncLogger(scf, log_config) as logger:
        for log_entry in logger:
            data = log_entry[1]

            var_x_history.append(data['kalman.varPX'])
            var_x_history.pop(0)
            var_y_history.append(data['kalman.varPY'])
            var_y_history.pop(0)
            var_z_history.append(data['kalman.varPZ'])
            var_z_history.pop(0)

            min_x = min(var_x_history)
            max_x = max(var_x_history)
            min_y = min(var_y_history)
            max_y = max(var_y_history)
            min_z = min(var_z_history)
            max_z = max(var_z_history)

            if (max_x - min_x) < threshold and (
                    max_y - min_y) < threshold and (
                    max_z - min_z) < threshold:
                break


__version__ = '1.0.1'

from .control import Controller

from .crazy import CrazyDragon

from .guidance import Guidance

from .guidance import utils

from .navigation import Navigation

from .communication_test import *


from navigation import Navigation

from control import Controller

from cflib.crazyflie.syncCrazyflie import SyncCrazyflie
from cflib.crazyflie               import Crazyflie
from cflib.utils                   import uri_helper

from numpy import array



nav_config = {
    'header'   : array([ 4,26]).tobytes(),
    'scf'      : None,
    'body_name': 'cf1',
    'port'     : '/dev/ttyUSB0',
    'baud'     : 115200
}

ctr_config = {
    'header': array([26, 4]).tobytes(),
    'scf'   : None,
    'dt'    : 0.1,
    'Hz'    : 5,
    'port'  : '/dev/ttyUSB0',
    'baud'  : 115200
}

uri1 = uri_helper.uri_from_env(default='radio://0/65/2M/E7E7E7E707')


if __name__ == "__main__":

    with SyncCrazyflie( uri1, cf=Crazyflie( rw_cache='./cache' ) ) as scf:

        nav_config['scf'] = scf
        ctr_config['scf'] = scf

        NAV = Navigation( nav_config )
        CTR = Controller( ctr_config )

        NAV.start()
        CTR.start()
